/*
  Running Win32 application as soon as possible after Windows Nt boot.

  0. Compile your program with included source code or #define W32ASAP_EXE and compile W32ASAP.exe.
     Make sure Subsystem in exe's IMAGE_OPTIONAL_HEADER is GUI (2) (link -edit W32ASAP.exe /subsystem:WINDOWS).
  1. Add REG_EXPAND_SZ value named, for example, "W32ASAP" to
     "HKLM\System\CurrentControlSet\Control\Session Manager\SubSystems".
     Set contents of the "W32ASAP" value to "FullPath\W32ASAP.exe Parameters". 
  2. Add "W32ASAP" as the last token (must be somewhere after "Windows" token) to contents of value
     "HKLM\System\CurrentControlSet\Control\Session Manager\SubSystems\Required".
  3. Restart the machine.

  A. W32ASAP process is created anytime a session is created before wininit/winlogon
     and sevice processes and runs in LocalSystem with no window station associated.
  B. W32ASAP image format must be the same as the native image format for given platform.
  C. The technique works on any Nt version.
*/

/* =========================== Required =========================== */
#ifndef _W32ASAP_H_
#define _W32ASAP_H_
#define WIN32_LEAN_AND_MEAN

#include <stdio.h>
#include <windows.h>

typedef LONG NTSTATUS;

typedef struct _UNICODE_STRING {
  USHORT Length;
  USHORT MaximumLength;
  PWSTR  Buffer;
} UNICODE_STRING, *PUNICODE_STRING;

typedef struct _OBJECT_ATTRIBUTES {
  ULONG Length;
  HANDLE RootDirectory;
  PUNICODE_STRING ObjectName;
  ULONG Attributes;
  PVOID SecurityDescriptor;
  PVOID SecurityQualityOfService;
} OBJECT_ATTRIBUTES, *POBJECT_ATTRIBUTES;

typedef struct _PORT_MESSAGE64 {
  USHORT DataLength;
  USHORT TotalLength;
  ULONG  NotImportant[9];
} PORT_MESSAGE64, *PPORT_MESSAGE64;

#pragma comment(lib, "ntdll.lib")
#define APILINK __declspec(dllimport)

APILINK NTSTATUS WINAPI ZwCreatePort(
  PHANDLE PortHandle,
  POBJECT_ATTRIBUTES ObjectAttributes,
  ULONG MaxConnectionInfoLength,
  ULONG MaxMessageLength,
  ULONG MaxPoolUsage
);

APILINK NTSTATUS WINAPI ZwConnectPort(
  PHANDLE PortHandle,
  PUNICODE_STRING PortName,
  PSECURITY_QUALITY_OF_SERVICE SecurityQos,
  PVOID ClientView,
  PVOID ServerView,
  PULONG MaxMessageLength,
  PVOID ConnectionInformation,
  PULONG ConnectionInformationLength
);

APILINK NTSTATUS WINAPI ZwListenPort(
  HANDLE PortHandle,
  PVOID RequestMessage
);

APILINK NTSTATUS WINAPI ZwAcceptConnectPort(
  PHANDLE PortHandle,
  PVOID PortContext,
  PVOID ConnectionRequest,
  BOOLEAN AcceptConnection,
  PVOID ServerView,
  PVOID ClientView
);

APILINK NTSTATUS WINAPI ZwCompleteConnectPort(
  HANDLE PortHandle
);

APILINK NTSTATUS WINAPI ZwAllocateLocallyUniqueId(
  PLUID pLuid
);

#undef APILINK

typedef struct _SB_CONNECTION_REQUEST {
  ULONG SbImageType; /* Windows - 2, OS/2 - 5, POSIX - 7 */
  WCHAR SbApiPortName[120];
} SB_CONNECTION_REQUEST, *PSB_CONNECTION_REQUEST;

typedef NTSTATUS (WINAPI *TRtlConnectToSm)(
  PUNICODE_STRING SbApiPortName,
  HANDLE SbbApiPort, /* if SbApiPortName present, checked for != 0 only */
  ULONG SbImageType, /* if SbApiPortName present, checked for != 0 and registered; must be != 0xFFFFFFFF */
  PHANDLE pSmApiPort
);

DWORD GetUid(VOID) {
  for(;;) {
    LUID Luid;
    DWORD Uid;
    ZwAllocateLocallyUniqueId(&Luid);
    Uid = Luid.LowPart + Luid.HighPart;
    if((Uid > 0x10) && (Uid != 0xFFFFFFFF)) {
      return(Uid);
    }
  } 
}

NTSTATUS WINAPI SmSsContinue(VOID) {
  OBJECT_ATTRIBUTES oa = {sizeof(oa)};
  UNICODE_STRING SbApiPortName;
  DWORD Uid;
  HANDLE SbApiPort;
  NTSTATUS NtStatus;
  WCHAR szSbApiPortName[64];
  Uid = GetUid();
  #define SbApiPortNamePfx L"\\W32ASAPPort"
  SbApiPortName.Length = (USHORT)(2*swprintf(SbApiPortName.Buffer = szSbApiPortName, SbApiPortNamePfx L"%u", Uid));
  SbApiPortName.MaximumLength = SbApiPortName.Length + 2;
  #undef SbApiPortNamePfx
  oa.ObjectName = &SbApiPortName;
  NtStatus = ZwCreatePort(&SbApiPort, &oa, 0, sizeof(PORT_MESSAGE64), 0);
  if(NtStatus >= 0) {
    HANDLE SmApiPort;
    TRtlConnectToSm pRtlConnectToSm = (TRtlConnectToSm)GetProcAddress(GetModuleHandleW(L"ntdll.dll"), "RtlConnectToSm");
    if(pRtlConnectToSm) {
      NtStatus = pRtlConnectToSm(&SbApiPortName, SbApiPort, Uid, &SmApiPort);
    }
    else {
      SECURITY_QUALITY_OF_SERVICE QoS = {sizeof(QoS), DEFAULT_IMPERSONATION_LEVEL, SECURITY_DYNAMIC_TRACKING, TRUE};
      UNICODE_STRING ApiPortName;
      SB_CONNECTION_REQUEST SbConnectionRequest;
      ULONG SbConnectionRequestSize;
      SbConnectionRequest.SbImageType = Uid;
      memcpy(SbConnectionRequest.SbApiPortName, szSbApiPortName, SbApiPortName.MaximumLength);
      ApiPortName.MaximumLength = sizeof(L"\\SmApiPort");
      ApiPortName.Length =  ApiPortName.MaximumLength - 2;
      ApiPortName.Buffer =  L"\\SmApiPort";
      SbConnectionRequestSize = sizeof(SbConnectionRequest);
      NtStatus = ZwConnectPort(&SmApiPort, &ApiPortName, &QoS, NULL, NULL, NULL, &SbConnectionRequest, &SbConnectionRequestSize);
    }
    if(NtStatus >= 0) {
      PORT_MESSAGE64 Msg;
      Msg.DataLength = 0;
      Msg.TotalLength = sizeof(Msg);
      if(ZwListenPort(SbApiPort, &Msg) >= 0) {
        HANDLE hPort;
        if(ZwAcceptConnectPort(&hPort, NULL, &Msg, TRUE, NULL, NULL) >= 0) {
          ZwCompleteConnectPort(hPort);
          CloseHandle(hPort);
        }
      }
      CloseHandle(SmApiPort);
    }
    CloseHandle(SbApiPort);
  }
  return(NtStatus);
}
#endif /* _W32ASAP_H_ */


#ifdef W32ASAP_EXE
/* =========================== Utils =========================== */
#ifdef _WIN64
  #define TEB_PEB               0x0060
  #define PEB_NT5_SESSION_ID    0x02C0
#else
  #define TEB_PEB               0x0030
  #define PEB_NT5_SESSION_ID    0x01D4
#endif
#define NtCurrentPeb() (*(PBYTE*)(((PBYTE)NtCurrentTeb())+TEB_PEB))
#define PEB32_NT4_SESSION_ID    0x014C

#define UserSharedData() ((volatile BYTE*)0x7FFE0000)
#define KUSER_TICK_COUNT        0x0000
#define KUSER_IMAGE_NUMBER_LOW  0x002C
#define KUSER_SYSTEM_ROOT       0x0030
#define KUSER_NT_MAJOR_VERSION  0x026C
#define KUSER_ACTIVE_CONSOLE_ID 0x02D8
#define KUSER_TICK_COUNT_NEW    0x0320

#define GetNtMajorVersion() (*(PBYTE)(UserSharedData()+KUSER_NT_MAJOR_VERSION)) /* NT3 == 0 */
#define GetSystemRootDir() ((PWSTR)(UserSharedData()+KUSER_SYSTEM_ROOT))
#define GetConsoleSessionId() (*(PULONG)(UserSharedData()+KUSER_ACTIVE_CONSOLE_ID))
#define AmIi386() (*(PWORD)(UserSharedData()+KUSER_IMAGE_NUMBER_LOW) == IMAGE_FILE_MACHINE_I386)
#ifdef _WIN64
  #define AmIWow64() 0
  #define GetCurrentSessionId() (*(PULONG)(NtCurrentPeb()+(PEB_NT5_SESSION_ID)))
#else
  #define AmIWow64() (!AmIi386())
  #define GetCurrentSessionId() (*(PULONG)(NtCurrentPeb()+(GetNtMajorVersion() > 4 ? PEB_NT5_SESSION_ID : PEB32_NT4_SESSION_ID)))
#endif

/* =========================== Bonus =========================== */
DWORD WINAPI WriteLog(const char *Text, DWORD nBytes) {
  HANDLE hFile;
  DWORD wBytes = 0;
  WCHAR LogNameW[64];
  swprintf(LogNameW, L"%lc:\\W32ASAP.log", GetSystemRootDir()[0]);
  hFile = CreateFileW(LogNameW, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, 0, NULL);
  if(hFile != INVALID_HANDLE_VALUE) {
    SetFilePointer(hFile, 0, NULL, FILE_END);
    WriteFile(hFile, Text, nBytes, &wBytes, NULL);
    CloseHandle(hFile);
  }
  return(wBytes);
}

__declspec(dllimport) int __cdecl DbgPrint(const char*, ...);

DWORD WINAPI FastTasks(INT argc, WCHAR *wargv[]) {
  SYSTEMTIME lt;
  DWORD n;
  CHAR Text[80];
  if(argc && wargv[0]) { /* just to reference argc, wargv */
    DbgPrint("%ls\n", GetCommandLineW());
  }
  GetLocalTime(&lt);
  n = (DWORD)sprintf(Text, "%02u %08X %u-%02u-%02u %02u:%02u:%02u.%03u\r\n",
    GetCurrentSessionId(), GetTickCount(),
    lt.wYear, lt.wMonth, lt.wDay, lt.wHour, lt.wMinute, lt.wSecond, lt.wMilliseconds
  );
  if(n) {
    n = WriteLog(Text, n);
  }
  return(n);
}

__declspec(dllimport) NTSTATUS WINAPI ZwOpenEvent(PHANDLE, DWORD, POBJECT_ATTRIBUTES);
__declspec(dllimport) NTSTATUS WINAPI ZwQuerySystemInformation(ULONG, PVOID, ULONG, PULONG);

DWORD WINAPI SlowTasks(INT argc, WCHAR *wargv[]) {
  DWORD n = 0;
  /* To do things once use better methods than this: */
  if((GetCurrentSessionId() == 0) /*&& GetConsoleSessionId() == 0)*/ && (GetTickCount() < 0x20000)) {
    if(argc > 1) {
      STARTUPINFOW si = {sizeof(si)};
      PROCESS_INFORMATION pi;
#ifdef _WIN64
      LPCWSTR WinStaName = L"Service-0x0-3e7$";
#else
      LPCWSTR WinStaName = GetNtMajorVersion() >= 4 ? L"Service-0x0-3e7$" : L"WinSta0";
#endif
      LPCWSTR DeskTpName = L"Default";
      CHAR Text[160];
#ifdef ALLOW_USER32
      HWINSTA hWinSta = NULL;
#else
      BYTE Found;
#ifdef SUPPORT_NT3
      BYTE b[0x4000];
      DWORD CurrentPid = GetCurrentProcessId();
#else
      UNICODE_STRING WinStaNameUS;
      OBJECT_ATTRIBUTES oa = {sizeof(oa)};
      HANDLE hEvent;
      oa.ObjectName = &WinStaNameUS;
#endif /* SUPPORT_NT3 */
#endif /* ALLOW_USER32 */
      if(argc > 2) {
        WinStaName = L"WinSta0";
        if(wargv[2][0] == L'0') {
          DeskTpName = L"WinLogon";
        }
      }
#if !defined(SUPPORT_NT3) && !defined(ALLOW_USER32)
      WinStaNameUS.Length = (USHORT)(2*swprintf(WinStaNameUS.Buffer = (PWSTR)Text, L"\\Windows\\WindowStations\\%ls", WinStaName));
      WinStaNameUS.MaximumLength = WinStaNameUS.Length + 2;
#endif
      do {
        Sleep(64);
#ifdef ALLOW_USER32
        hWinSta = OpenWindowStationW(WinStaName, FALSE, READ_CONTROL);
        if(hWinSta != NULL) {
          break;
        }
#else
#ifndef SUPPORT_NT3
        /* When linking with user32.dll is not suitable (~ no AppInit_DLLs),
           check if requested WinSta exists this way: */
        #ifndef STATUS_OBJECT_TYPE_MISMATCH
          #define STATUS_OBJECT_TYPE_MISMATCH ((NTSTATUS)0xC0000024L)
        #endif
        if(ZwOpenEvent(&hEvent, READ_CONTROL, &oa) == STATUS_OBJECT_TYPE_MISMATCH) {
#else
        /* When linking with user32.dll is not suitable and you need to support NT3
           or to wait that services are ready, check if services.exe process exists. */
        Found = 0;
        if(ZwQuerySystemInformation(5, b, sizeof(b), NULL) >= 0) {
          typedef struct _SYSTEM_PROCESSES {
            ULONG NextEntryOffset;
            ULONG NotImportant0[13];
            UNICODE_STRING ProcessName;
            ULONG_PTR Reserved1;
            ULONG_PTR Pid;
          } *PSYSTEM_PROCESSES;
          PSYSTEM_PROCESSES p = (PSYSTEM_PROCESSES)b;
          do {
            if(Found & 1) {
              if(p->ProcessName.Length == sizeof(L"services.ex")) {
                if(!wcsicmp(L"services.exe", p->ProcessName.Buffer)) {
                  Found |= 2;
                  break;
                }
              }
            }
            else {
              if(p->Pid == CurrentPid) {
                Found |= 1;
              }
            }
            if(p->NextEntryOffset == 0) {
              break;
            }
            p = (PSYSTEM_PROCESSES)((PBYTE)p + p->NextEntryOffset);
          } while(!(Found & 2));
          if(Found & 2) {
            break;
          }
        }
        else {
#endif /* SUPPORT_NT3 */
          Found |= 2;
          break;
        }
#endif /* ALLOW_USER32 */
      } while(++n < 1000);
#ifdef ALLOW_USER32
      if(hWinSta) {
        CloseWindowStation(hWinSta);
#else
      if(Found & 2) {
#endif /* ALLOW_USER32 */
        swprintf(si.lpDesktop = (LPWSTR)Text, L"%ls\\%ls", WinStaName, DeskTpName);
        n = 0;
        /* Can be 32bit image, can be console app. Test what can be spawned. */
        if(CreateProcessW(NULL, wargv[1], NULL, NULL, 0, FALSE, NULL, NULL, &si, &pi)) {
          CloseHandle(pi.hThread);
          if(WaitForSingleObject(pi.hProcess, 2000) == STATUS_WAIT_0) {
            GetExitCodeProcess(pi.hProcess, &n);
            n = (DWORD)sprintf(Text, "Child process '%.80ls' exited (%u).\r\n", wargv[1], n);
          }
          CloseHandle(pi.hProcess);
        }
        else {
          n = (DWORD)sprintf(Text, "CreateProcess failed (%u).\r\n", GetLastError());
        }
      }
      else {
        n = (DWORD)sprintf(Text, "FindWinSta failed (%u).\r\n", GetLastError());
      }
      if(n) {
        n = WriteLog(Text, n);
      }
    }
  }
  return(!n);
}

#ifndef USE_CRTLIB
INT xwmain(INT argc, WCHAR *wargv[]) {
#else
INT wmain(INT argc, WCHAR *wargv[]) {
#endif
  if((LONG)GetVersion() < 0) {
    return(0);
  }
  SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOALIGNMENTFAULTEXCEPT | SEM_NOOPENFILEERRORBOX);
  FastTasks(argc, wargv);
  SmSsContinue();
  return((INT)SlowTasks(argc, wargv));
}

#ifndef USE_CRTLIB
INT WINAPI ParseCommandLineW(LPCWSTR lpInput, LPWSTR *argv, INT nargv, LPWSTR lpCommandLine, ULONG ccharCommandLine) {
  INT argc = 0;
  LPWSTR c = lpCommandLine;
  LPCWSTR p = lpInput;
  WCHAR s;
  while(ccharCommandLine-- && ((*c++ = *p++) != 0));
  if((ccharCommandLine == 0) && *p) {
    return(0);
  }
  c = lpCommandLine;
  for(;;) {
    while(*c && (*c <= ' ')) {
      c++;
    }
    if(*c == '\0') {
      break;
    }
    s = *c == '"' ? *c++ : ' ';
    if(argc < nargv) {
      argv[argc++] = c;
    }
    while(*c && (*c != s)) {
      c++;
    }
    if(*c == '\0') {
      break;
    }
    *c++ = '\0';
  };
  return(argc);
}

INT mainCRTStartup(VOID) {
  WCHAR CommandLineW[0x7D0];
  LPWSTR wargv[3];
  ExitProcess((DWORD)xwmain(ParseCommandLineW(GetCommandLineW(), wargv, sizeof(wargv)/sizeof(*wargv), CommandLineW, sizeof(CommandLineW)/2), wargv));
}
#endif /* !USE_CRTLIB */
#endif /* W32ASAP_EXE */
